
import std.stdio;
import std.format;
import std.string;

import core.sys.posix.signal;
import core.sys.posix.unistd;
import core.sys.posix.sys.ioctl;
//import core.sys.freebsd.config;
import core.sys.freebsd.sys.event;
import core.stdc.errno;

import ecap, edsm;

class EventSource {
    EventQueue eq;
    StageMachine owner;
    char tag;               /* 'D', 'T', 'S', 'L' etc */

    int id;         		/* fd (usually, but not always) */
    short evtype;		    /* -> kevent.filter */
    intptr_t evdata;	    /* -> kevent.fdata */

    this(char tag) {
        this.tag = tag;
        eq = EventQueue.getInstance();
    }

    ~this() {
        // writefln("   !!! %s() : %s (owner %s, id = %d) this @ 0x%x", __FUNCTION__, this, owner.name, id, cast(void*)this);
    }

    void enable() {
        eq.enableEvent(evtype, this);
        //writefln("'%s' enabled %d/%d (%s)", owner.name, id, evtype, this);
    }

    void disable() {
        eq.disableEvent(evtype, this);
        //writefln("'%s' disabled %d/%d (%s)", owner.name, id, evtype, this);
    }

    abstract ulong eventCode(kevent_t ke);
}

final class Signal : EventSource {

    enum int sigInt = SIGINT;
    enum int sigTerm = SIGTERM;
    ulong number;

    this(int signo) {
        super('S');
        id = signo;
        evtype = EVFILT_SIGNAL;

        sigset_t sset;
        sigset_t old_sset;
        /* block the signal */
        sigemptyset(&sset);
        sigaddset(&sset, signo);
        sigprocmask(SIG_BLOCK, &sset, &old_sset);
    }

    override ulong eventCode(kevent_t ke) {
        return number;
    }
}

final class Timer : EventSource {
    ulong nexp;     // >1 => overrun
    ulong number;

    this(int id) {
        super('T');
        evtype = EVFILT_TIMER;
        this.id = id;
    }

    void start(uint interval_ms) {
        evdata = interval_ms;
        enable();
    }

    void stop() {
        disable();
    }

    override ulong eventCode(kevent_t ke) {
        return number;
    }
}

class Io : EventSource {

    private long _bytesAvail;

    this() {
        super('D');
        id = -1;
    }

    this(int fd) {
        super('D');
        id = fd;
    }

    long bytesAvail() {
        return _bytesAvail;
    }

    void enable(ushort evtype) {
        eq.enableEvent(evtype, this);
        //writefln("'%s' enabled %d/%d (%s)", owner.name, id, evtype, this);
    }

    void enableCanRead() {
        eq.enableEvent(EVFILT_READ, this);
        //writefln("'%s' enabled %d (%s)", owner.name, id, this);
    }

    void enableCanWrite() {
        eq.enableEvent(EVFILT_WRITE, this);
        //writefln("'%s' enabled %d (%s)", owner.name, id, this);
    }

    override ulong eventCode(kevent_t ke) {

        if (EVFILT_READ == ke.filter) {
            _bytesAvail = ke.data;
            return 0;
        }
        if (EVFILT_WRITE == ke.filter) {
            return 1;
        }

        writefln("%s(): ke.filter = %d", __FUNCTION__, ke.filter);
        assert(false, "oops, unexpected evtype (filter)");
    }
}

final class ClientSocket : Io {

    import core.sys.posix.sys.socket;
    import core.sys.posix.netinet.in_;
    import core.sys.posix.netdb;
    import core.stdc.string;
    import core.sys.posix.fcntl;

    string host;
    string port;
    addrinfo hints;
    addrinfo *ai;

    this(string host, string port) {
        super();
        this.host = host;
        this.port = port;
        hints.ai_family = AF_INET;
        hints.ai_socktype = SOCK_STREAM;
    }

    void getId() {
        getaddrinfo(toStringz(host), toStringz(port), &hints, &ai);
        id = socket(ai.ai_family, ai.ai_socktype, ai.ai_protocol);
        assert(id > 0, format("socket() failed : %s", fromStringz(strerror(errno))));
    }

    bool startConnect() {

        int flags = fcntl(id, F_GETFL);
        flags |= O_NONBLOCK;
        fcntl(id, F_SETFL, flags);

        errno = 0;
        int r = connect(id, ai.ai_addr, ai.ai_addrlen);
        if (-1 == r) {

            if (EINPROGRESS == errno)
                goto __ok;

            writefln("%s() : connect() : %s", __FUNCTION__, fromStringz(strerror(errno)));
            return false;
        }

      __ok:
        errno = 0;
        return true;
    }

    bool connOk() {
        /*
        tcp_info tcpi; // linux specific, see /usr/include/linux/tcp.h
        socklen_t tcpi_len = tcpi.sizeof;

        getsockopt(id, SOL_TCP, TCP_INFO, &tcpi, &tcpi_len);
        if (tcpi.tcpi_state != TCP_ESTABLISHED)
            return false;
        */
        return true; // TODO
    }

    void putId() {
        freeaddrinfo(ai);
        close(id);
    }

    int getError() {
        int e;
        socklen_t optlen = e.sizeof;
        getsockopt(id, SOL_SOCKET, SO_ERROR, &e, &optlen);
        return e;
    }
}

class Client {
    int fd;
    string addr;
    ushort port;

    this(int fd, string addr, ushort port) {
        this.fd = fd;
        this.addr = addr;
        this.port = port;
    }

    ~this() {
        writeln(__FUNCTION__);
        close(fd);
    }
}

final class TCPListener : EventSource {

    import core.sys.posix.sys.socket;
    import core.sys.posix.netinet.in_;
    import core.stdc.string;

    this(ushort port) {
        super('L');
        int sk = socket(AF_INET, SOCK_STREAM, 0);
        assert(sk > 0, format("socket() failed"));
        int on = 1;
        setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &on, on.sizeof);

        sockaddr_in myAddr;
        myAddr.sin_family = AF_INET;
        myAddr.sin_port = htons(port);
        bind(sk, cast(sockaddr*)&myAddr, myAddr.sizeof);
        listen(sk, 128);
        id = sk;
        evtype = EVFILT_READ;
    }

    Client acceptClient() {
        sockaddr_in peerAddr;
        socklen_t addrLength = peerAddr.sizeof;
        char[32] addr;
        int sk = accept(id, cast(sockaddr*)&peerAddr, &addrLength);
        inet_ntop(AF_INET, &peerAddr.sin_addr, addr.ptr, cast(uint)addr.length);
        return new Client(sk, addr[0 .. strlen(addr.ptr)].idup, peerAddr.sin_port);
    }

    override ulong eventCode(kevent_t ke) {
        return 0;
    }
}
